type 'a t = ('a * 'a t lazy_t) option

let id x = x

let init f = 
  let rec loop i = 
    let v, e = try (f i, false) with Not_found -> (f 0, true) in
    if e then None
    else Some(v, lazy(loop (i+1)))
  in
  try
    Some(f 0, lazy(loop 1))
  with Not_found -> None

let rec of_list = function
  | [] -> None
  | x::xs -> Some(x, lazy(of_list xs))

let of_array ary =
  let rec loop i =
    if i >= Array.length ary then None
    else Some(ary.(i), lazy(loop (i+1)))
  in
  loop 0

let empty = None

let cons v l = Some(v, lazy l)
let lcons v l = Some(v, l)

let numbers x =
  let rec loop i = Some(i, lazy(loop (i+1))) in
  loop x

let natural = numbers 0

let from f =
  let rec loop () =
    try
      Some(f(), lazy(loop ()))
    with Not_found -> None
  in
  loop()

let rec repeat x = Some(x, lazy(repeat x))
let cycle l =
  let rec loop = function
    | None -> loop l
    | Some(h, t) -> Some(h, lazy(loop (Lazy.force t)))
  in
  loop l

let rec unfoldr f s =
  match f s with
  | Some(v, s') -> Some(v, lazy(unfoldr f s'))
  | None -> None

let single x = Some(x, lazy None) 
let is_single = function
  | None -> false
  | Some(_, xs) -> (Lazy.force xs) = None
let wrap = single

let of_string str =
  let rec loop i =
    if i >= String.length str then None
    else Some(str.[i], lazy(loop (i+1)))
  in
  loop 0

let split_string c str =
  let rec loop s i =
    if i >= String.length str then Some(String.sub str s (i-s), lazy(None))
    else if str.[i] = c then
      if s = i then loop (s+1) (s+1)
      else Some(String.sub str s (i-s), lazy(loop (i+1) (i+1)))
    else loop s (i+1)
  in
  loop 0 0

let of_lines = split_string '\n'
let of_words = split_string ' '

let rec iter f = function
  | None -> ()
  | Some(x, xs) -> ignore(f x); iter f (Lazy.force xs)
let rec map f = function
  | None -> None
  | Some(x, xs) -> Some(f x, lazy(map f (Lazy.force xs)))
let rec fold_left f res = function
  | None -> res
  | Some(x, xs) -> fold_left f (f res x) (Lazy.force xs)

let rec fold_right f s = function
  | None -> s
  | Some(x, xs) -> f x (fold_right f s (Lazy.force xs))

let rec iter2 f l1 l2 =
  match l1, l2 with
  | None, _ | _, None -> ()
  | Some(x, xs), Some(y, ys) ->
      ignore(f x y);
      iter2 f (Lazy.force xs) (Lazy.force ys)


let length l = fold_left (fun res _ -> res + 1) 0 l

let hd = function
  | Some(x, _) -> x
  | None -> failwith "hd"

let tl = function
  | Some(_, xs) -> Lazy.force xs
  | None -> failwith "tl"

let pair = function
  | Some(x, xs) -> (x, Lazy.force xs)
  | None -> failwith "pair"

let force l =
  let rec loop = function
    | None -> ()
    | Some(_, xs) -> loop (Lazy.force xs)
  in
  loop l;
  l


let rec nth l n =
  match l, n with
  | None, _ -> raise Not_found
  | Some(x, xs), n ->
      if n = 0 then x
      else nth (Lazy.force xs) (n-1)

let (@@) = nth

let rev l =
  let rec loop res = function
    | None -> res
    | Some(x, xs) -> loop (Some(x, lazy res)) (Lazy.force xs)
  in
  loop None l


let rec append l1 l2 =
  match l1, l2 with
  | None, _ -> l2
  | _, None -> l1
  | Some(x, xs), Some(_, _) -> Some(x, lazy(append (Lazy.force xs) l2))

let (+@) = append


let rec concat = function
  | None -> None
  | Some(l, ls) -> append l (concat (Lazy.force ls))

let flatten = concat



let rec map2 f l1 l2 = 
  match l1, l2 with
  | None, _  | _, None -> None
  | Some(x, xs), Some(y, ys) ->
      Some(f x y, lazy(map2 f (Lazy.force xs) (Lazy.force ys)))

let rec fold_left2 f s l1 l2 =
  match l1, l2 with
  | None, _ | _, None -> s
  | Some(x, xs), Some(y, ys) ->
      fold_left2 f (f s x y) (Lazy.force xs) (Lazy.force ys)
let rec fold_right2 f s l1 l2 =
  match l1, l2 with
  | None, _ | _, None -> s
  | Some(x, xs), Some(y, ys) ->
      f x y (fold_right2 f s (Lazy.force xs) (Lazy.force ys))

let rec for_all p = function
  | None -> true
  | Some(x, xs) -> if p x then for_all p (Lazy.force xs) else false
let rec exists p = function
  | None -> false
  | Some(x, xs) -> if p x then true else exists p (Lazy.force xs)

let all_and = for_all id
let all_or = exists id
let sum = fold_left (+) 0
let product = fold_left ( * ) 1
let max ?(cmp=compare) =
  let rec loop maxval = function
    | None -> maxval
    | Some(x, xs) -> match cmp maxval x with
      | minus when minus < 0 -> loop x (Lazy.force xs)
      | _ -> loop maxval (Lazy.force xs)
  in
  function
    | None -> raise Not_found
    | Some(x, xs) -> loop x (Lazy.force xs)

let min ?(cmp=compare) = max ~cmp:(fun x y -> -1 * cmp x y)


let rec mem v = function
  | None -> false
  | Some(x, _) when v = x -> true
  | Some(_, xs) -> mem v (Lazy.force xs)
let is_empty = function
  | None -> true
  | _ -> false


let rec find p = function
  | None -> raise Not_found
  | Some(x, _) when p x -> x
  | Some(_, xs) -> find p (Lazy.force xs)
let find_index p =
  let rec loop i = function
    | None -> raise Not_found
    | Some(x, _) when p x -> i
    | Some(_, xs) -> loop (i+1) (Lazy.force xs)
  in
  loop 0

let rec filter p = function
  | None -> None
  | Some(x, xs) when p x -> Some(x, lazy(filter p (Lazy.force xs)))
  | Some(_, xs) -> filter p (Lazy.force xs)
let find_all = filter
let find_indices p =
  let rec loop i = function
    | None -> None
    | Some(x, xs) when p x -> Some(i, lazy(loop (i+1) (Lazy.force xs)))
    | Some(_, xs) -> loop (i+1) (Lazy.force xs)
  in
  loop 0
let partition p =
  let rec loop l1 l2 = function
    | None -> rev l1, rev l2
    | Some(x, xs) ->
        if p x then loop (Some(x, lazy l1)) l2 (Lazy.force xs)
        else loop l1 (Some(x, lazy l2)) (Lazy.force xs)
  in
  loop None None
let rec unique ?(cmp=(=)) = function
  | None -> None
  | Some(x, xs) ->
      let res = filter (fun y -> not (cmp x y)) (Lazy.force xs) in
      Some(x, lazy(unique ~cmp (Lazy.force xs)))
  
let rec take n = function
  | Some(x, xs) when n != 0 -> Some(x, lazy(take (n-1) (Lazy.force xs)))
  | _ -> None

let rec drop n = function
  | Some(_, xs) when n != 0 -> drop (n-1) (Lazy.force xs)
  | l -> l
let rec split_at n l =
  (take n l, drop n l)

let rec take_while p = function
  | Some(x, xs) when p x -> Some(x, lazy(take_while p (Lazy.force xs)))
  | _ -> None

let rec drop_while p = function
  | Some(x, xs) when p x -> drop_while p (Lazy.force xs)
  | l -> l
let rec span p l = (take_while p l, drop_while p l)

let group ?(cmp=(=)) =
  let rec loop res = function
    | None -> single res
    | Some(x, xs) ->
        if x = hd res then loop (Some(x, lazy res)) (Lazy.force xs)
        else Some(res, lazy(loop (single x) (Lazy.force xs)))
  in
  function
    | None -> None
    | Some(x, xs) -> loop (single x) (Lazy.force xs)


let rec assoc k = function
  | None -> raise Not_found
  | Some((k', v'), _) when k = k' -> v'
  | Some(_, xs) -> assoc k (Lazy.force xs)

let rec mem_assoc k = function
  | None -> false
  | Some((k', _), _) when k = k' -> true
  | Some(_, xs) -> mem_assoc k (Lazy.force xs)

let rec filter_assoc k = function
  | None -> None
  | Some((k', _) as x, xs) ->
      if k = k' then Some(x, lazy(filter_assoc k (Lazy.force xs)))
      else filter_assoc k (Lazy.force xs)


let rec combine l1 l2 =
  try
    let (h1, t1), (h2, t2) = pair l1, pair l2 in
    Some((h1, h2), lazy(combine t1 t2))
  with Failure _ -> None
let split l =
  let rec loop l1 l2 = function
    | None -> (rev l1, rev l2)
    | Some((x1, x2), xs) ->
        loop (Some(x1, lazy l1)) (Some(x2, lazy l2)) (Lazy.force xs)
  in
  loop None None l

let zip = combine
let rec zip3 l1 l2 l3 =
  try
    let (h1, t1), (h2, t2), (h3, t3) = pair l1, pair l2, pair l3 in
    Some((h1, h2, h3), lazy(zip3 t1 t2 t3))
  with Failure _ -> None

let rec zip4 l1 l2 l3 l4 =
  try
    let ((h1, t1), (h2, t2), (h3, t3), (h4, t4)) =
      pair l1, pair l2, pair l3, pair l4 in
    Some((h1, h2, h3, h4), lazy(zip4 t1 t2 t3 t4))
  with Failure _ -> None

let unzip = split
let unzip3 l =
  let rec loop l1 l2 l3 = function
    | None -> (rev l1, rev l2, rev l3)
    | Some((x1, x2, x3), xs) ->
        loop (Some(x1, lazy l1)) (Some(x2, lazy l2)) (Some(x3, lazy l3))
          (Lazy.force xs)
  in
  loop None None None l
let unzip4 l =
  let rec loop l1 l2 l3 l4 = function
    | None -> (rev l1, rev l2, rev l3, rev l4)
    | Some((x1, x2, x3, x4), xs) ->
        loop (Some(x1, lazy l1)) (Some(x2, lazy l2))
          (Some(x3, lazy l3)) (Some(x4, lazy l4)) (Lazy.force xs)
  in
  loop None None None None l

let rec rev_append l = function
  | None -> l
  | Some(x, xs) -> rev_append (Some(x, lazy l)) (Lazy.force xs)


let rec merge ?(cmp=compare) l1 l2 =
  match l1, l2 with
  | None, _ -> l2
  | _, None -> l1
  | Some(x, xs), Some(y, ys) ->
      if cmp x y <= 0 then Some(x, lazy(merge ~cmp (Lazy.force xs) l2))
      else Some(y, lazy(merge ~cmp l1 (Lazy.force ys)))

let sort ?(cmp=compare) l =
  let rec merge_pairs = function
    | None -> None
    | Some(x, xs) as l ->
        if is_empty x then l
        else
          let x2, xss = pair (Lazy.force xs) in
          cons (merge ~cmp x x2) (merge_pairs xss)
  in
  let rec loop = function
    | None -> None
    | x when is_single x -> hd x
    | l -> loop (merge_pairs l)
  in
  loop (map single l)


let rec transpose = function
  | None -> None
  | Some(x, xs) when is_empty x -> transpose (Lazy.force xs)
  | Some(xs, xss) ->
      let x, xs' = pair xs in
      let hs, ts = unzip (map pair (Lazy.force xss)) in
      Some((cons x hs), lazy(transpose (cons xs' ts)))


let rec to_list = function
  | None -> []
  | Some(x, xs) -> x :: (to_list (Lazy.force xs))
let to_string l =
  let ret = String.create (length l) in
  let rec loop i = function
    | None -> ret
    | Some(x, xs) ->
        ret.[i] <- x;
        loop (i+1) (Lazy.force xs)
  in
  loop 0 l

