type 'a t = ('a * 'a t lazy_t) option

(** {6 constructors } *)

val init: (int -> 'a) -> 'a t
val of_list: 'a list -> 'a t
val of_array: 'a array -> 'a t
val empty: 'a t
val cons: 'a -> 'a t -> 'a t
val lcons: 'a -> 'a t lazy_t -> 'a t
val numbers: int -> int t
val natural: int t
val from: (unit -> 'a) -> 'a t
val repeat: 'a -> 'a t
val cycle: 'a t -> 'a t
val unfoldr: ('a -> ('b * 'a) option) -> 'a -> 'b t

val single: 'a -> 'a t
val is_single: 'a t -> bool
val wrap: 'a -> 'a t

val of_string: string -> char t
val of_lines: string -> string t
val of_words: string -> string t


val to_list: 'a t -> 'a list
val to_string: char t -> string

(** {6 accessors } *)

val length: 'a t -> int

val hd: 'a t -> 'a

val tl: 'a t -> 'a t

val pair: 'a t -> 'a * 'a t

val force: 'a t -> 'a t
(** [force l] returns l itself. This function does nothing without
forcing any lazy evaluations on the LazyList l *)


val nth: 'a t -> int -> 'a

val (@@): 'a t -> int -> 'a (* same as nth *)

val rev: 'a t -> 'a t

val append: 'a t -> 'a t -> 'a t

val (+@): 'a t -> 'a t -> 'a t (* same as append *)




val concat: 'a t t -> 'a t

val flatten: 'a t t -> 'a t (* same as concat *)


(** {6 iterators } *)
val iter: ('a -> 'b) -> 'a t -> unit
val map: ('a -> 'b) -> 'a t -> 'b t
val fold_left: ('a -> 'b -> 'a) -> 'a -> 'b t -> 'a
val fold_right: ('a -> 'b -> 'b) -> 'b -> 'a t -> 'b

(** {6 iterators on two lists } *)
val iter2: ('a -> 'b -> 'c) -> 'a t -> 'b t -> unit
val map2: ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
val fold_left2: ('a -> 'b -> 'c -> 'a) -> 'a -> 'b t -> 'c t -> 'a
val fold_right2: ('a -> 'b -> 'c -> 'c) -> 'c -> 'a t -> 'b t -> 'c

(** {6 special folds *} *)
val all_and: bool t -> bool
val all_or: bool t -> bool
val sum: int t -> int
val product: int t -> int
val max: ?cmp:('a -> 'a -> int) -> 'a t -> 'a
val min: ?cmp:('a -> 'a -> int) -> 'a t -> 'a


(** {6 predicates } *)
val for_all: ('a -> bool) -> 'a t -> bool
val exists: ('a -> bool) -> 'a t -> bool
val mem: 'a -> 'a t -> bool
val is_empty: 'a t -> bool


(** {6 filters } *)
val find: ('a -> bool) -> 'a t -> 'a
val find_index: ('a -> bool) -> 'a t -> int
val filter: ('a -> bool) -> 'a t -> 'a t
val find_all: ('a -> bool) -> 'a t -> 'a t
val find_indices: ('a -> bool) -> 'a t -> int t
val partition: ('a -> bool) -> 'a t -> 'a t * 'a t
val unique: ?cmp:('a -> 'a -> bool) -> 'a t -> 'a t

(** {6 Extracting sublists} *)
val take: int -> 'a t -> 'a t
val drop: int -> 'a t -> 'a t
val split_at: int -> 'a t -> ('a t * 'a t)

val take_while: ('a -> bool) -> 'a t -> 'a t
val drop_while: ('a -> bool) -> 'a t -> 'a t
val span: ('a -> bool) -> 'a t -> ('a t * 'a t)
val group: ?cmp:('a -> 'a -> bool) -> 'a t -> 'a t t


(** {6 association lists} *)
val assoc: 'a -> ('a * 'b) t -> 'b
val mem_assoc: 'a -> ('a * 'b) t -> bool
val filter_assoc: 'a -> ('a * 'b) t -> ('a * 'b) t


(** {6 Lists of pairs} *)
val combine: 'a t -> 'b t -> ('a * 'b) t
val split: ('a * 'b) t -> 'a t * 'b t

val zip: 'a t -> 'b t -> ('a * 'b) t (* same as combine *)
val zip3: 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
val zip4: 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t

val unzip: ('a * 'b) t -> 'a t * 'b t
val unzip3: ('a * 'b * 'c) t -> 'a t * 'b t * 'c t
val unzip4: ('a * 'b * 'c * 'd) t -> 'a t * 'b t * 'c t * 'd t


(** {6 sorting} *)
val sort: ?cmp:('a -> 'a -> int) -> 'a t -> 'a t
val merge: ?cmp:('a -> 'a -> int) -> 'a t -> 'a t -> 'a t



(** {6 functions from Haskell } *)
val transpose: 'a t t -> 'a t t

